//line 202 "Panning" replaces "noPanning", to solve the panning bug
//add "mod" pin in the processes:
//line 185
//line 194
//line 196
//line 219
//line 227
//line 279 end note buffer offset to -1, better stability with short asio buffers (2ms)
//and originally I modified this line to avoid clicks with no looped samples with the first version of the sampleoscillator2

#ifndef SoundfontOscillator3_H_INCLUDED
#define SoundfontOscillator3_H_INCLUDED

#include "mp_sdk_audio.h"
#include "soundfont_user.h"

class SoundfontOscillator3 : public MpBase, public soundfontUser
{
public:
	SoundfontOscillator3( IMpUnknown* host );
	~SoundfontOscillator3();
	virtual int32_t MP_STDCALL open();
	virtual void onSetPins(void);
	virtual void onGraphStart();

	IntInPin pinSampleId;
	EnumInPin pinQuality;
	AudioInPin pinPitch;
	AudioInPin pinTrigger;
	AudioInPin pinGate;
	AudioInPin pinVelocity;
	AudioOutPin pinLeft;
	AudioOutPin pinRight;
    IntInPin pinOn;//kx's change
	AudioInPin pinMod;//kx's change
	EnumInPin pinTracking;//kx's change


	void ChooseSubProcess( int blockPosistion );
	void process_with_gate( int bufferOffset, int sampleframes );

	//void sub_process_pitch_mono_hi_quality( int bufferOffset, int sampleframes );
	//void sub_process_pitch_mono_low_quality( int bufferOffset, int sampleframes);

	// Stereo, no pitch modulation.
	void sub_process_stereo_no_quality( int bufferOffset, int sampleframes );
	void sub_process_stereo_low_quality( int bufferOffset, int sampleframes );
	void sub_process_stereo_mid_quality( int bufferOffset, int sampleframes );
	void sub_process_stereo_hi_quality( int bufferOffset, int sampleframes );

	// Stereo, Panning support, no pitch modulation.
	void sub_process_stereo_panned_no_quality( int bufferOffset, int sampleframes );
	void sub_process_stereo_panned_low_quality( int bufferOffset, int sampleframes );
	void sub_process_stereo_panned_mid_quality( int bufferOffset, int sampleframes );
	void sub_process_stereo_panned_hi_quality( int bufferOffset, int sampleframes );


	// Stereo, Pitch modulation.
	void sub_process_pitch_stereo_no_quality( int bufferOffset, int sampleframes );
	void sub_process_pitch_stereo_hi_quality( int bufferOffset, int sampleframes );
	void sub_process_pitch_stereo_mid_quality( int bufferOffset, int sampleframes );
	void sub_process_pitch_stereo_low_quality( int bufferOffset, int sampleframes );

	// Stereo, Panning support, Pitch modulation.
	void sub_process_pitch_stereo_panned_no_quality( int bufferOffset, int sampleframes );
	void sub_process_pitch_stereo_panned_hi_quality( int bufferOffset, int sampleframes );
	void sub_process_pitch_stereo_panned_mid_quality( int bufferOffset, int sampleframes );
	void sub_process_pitch_stereo_panned_low_quality( int bufferOffset, int sampleframes );

	void sub_process_silence(int bufferOffset, int sampleframes);
	void NoteDone( int blockPosistion );
	void NoteOn( int blockPosistion );
	//void NoteOff( int blockPosistion );

	bool gate_state;

private:
	float* GetInterpolationtable();
	bool trigger_state;

	std::wstring FileName;

	SubProcess_ptr current_osc_func;
	bool sample_playing;

public:
	void ResetWave(void);
};

// 1-point.
class NoInterpolator
{
public:
	// needs 'Inline Function Expansion (Any Suitable)' enabled.
	inline static float Interpolate( short* const sample_ptr, int fine ) //const
	{
		const float scale = 1.f / (float) 0x10000; // shorts.
		return scale * sample_ptr[0];
	};
};

// 2-point.
class LinearInterpolator
{
public:
	// needs 'Inline Function Expansion (Any Suitable)' enabled.
	inline static float Interpolate( short* const sample_ptr, int fine ) //const
	{
		const float scale = ( 1.f / 0x10000 ) / (float) 0x1000000; // shorts.
		return ( scale *  ((float) sample_ptr[0] * ( 0x1000000 - (int)fine) + (float) sample_ptr[1] * (int)fine) );
	};
};

// 4-point
class CubicInterpolator
{
public:
	// needs 'Inline Function Expansion (Any Suitable)' enabled.
	inline static float Interpolate( short* const sample_ptr, int fine ) //const
	{
		const float scale = 1.f / (float) 0x10000; // shorts.

		float y0, y1, y2, y3, mu;
		y0 = (float) sample_ptr[-1];
		y1 = (float) sample_ptr[0];
		y2 = (float) sample_ptr[1];
		y3 = (float) sample_ptr[2];
		mu = (float)fine / (float) 0x1000000;

		float a0,a1,a2,a3,mu2;

		mu2 = mu*mu;
		a0 = y3 - y2 - y0 + y1;
		a1 = y0 - y1 - a0;
		a2 = y2 - y0;
		a3 = y1;

		return scale * ( a0 * mu * mu2 + a1 * mu2 + a2 * mu + a3 );
	};
};

// 7-point.
class SincInterpolator
{
public:
	// needs 'Inline Function Expansion (Any Suitable)' enabled.
	inline static float Interpolate( short* const sample_ptr, int fine ) //const
	{
		short* ptr = sample_ptr - 3;

		// reduce fraction part to range 0-31 then multiply by 8 (zeros lower 3 bits)
		float * interpolation_table_ptr = interpolation_table2 + (fine >> 19) * 8;

		// now look up interpolation filter table
		// perform 2 filters at once, one 1 sub sample ahead
		float a = 0.f;
		for( int x = 4 ; x > 0 ; x-- )
		{
			a += *ptr++ * *interpolation_table_ptr++;
			a += *ptr++ * *interpolation_table_ptr++; //unrolled
		}

		float a2 = 0.f;
		ptr = sample_ptr - 3;
		for( int x = 4 ; x > 0 ; x-- )
		{
			a2 += *ptr++ * *interpolation_table_ptr++;
			a2 += *ptr++ * *interpolation_table_ptr++;
		}

		// now linear interpolation between 2 subsamples
		const float scale = (2.f / 0x1000)/(float) 0x1000000; //shorts

		// save lowest 19 bits for interpolation
		int fract_part_fine = fine & 0x7ffff;
		return scale * (a * ( 0x80000 - fract_part_fine) + a2 * fract_part_fine);
	};

	static float* interpolation_table2;
};

class PitchFixed
{
public:
	inline static void CalculateIncrement( partial & p, float*& pitch, float*& mod )
	{
		// do nothing. Hopefully optimizes away to nothing.
	};
};

class PitchChanging
{
public:
	inline static void CalculateIncrement( partial& p, float*& pitch, float*& mod )
	{
		p.CalculateIncrement( *pitch++ + *mod++ );
	};
};

enum PanningSupport_t {Panning, noPanning };

template< class InterpolationPolicy, class PitchModulationPolicy = PitchFixed, PanningSupport_t PanningSupport = Panning >
class CodeGenerator : public InterpolationPolicy, private PitchModulationPolicy
{
public:

/*#if !defined( _DEBUG )
	__forceinline
#endif*/
	static void sub_process( SoundfontOscillator3& oscillator, int bufferOffset, int sampleframes ) //const
	{
		bool addToOutput = false;

		for( activePartialListType::iterator it = oscillator.partials.begin() ; it != oscillator.partials.end() ; )
		{
			partial &partial = *it;

			float* pitch    = oscillator.pinPitch.getBuffer() + bufferOffset;
            float* mod      = oscillator.pinMod.getBuffer() + bufferOffset;
			float* output_l = oscillator.pinLeft.getBuffer() + bufferOffset;
			float* output_r = oscillator.pinRight.getBuffer() + bufferOffset;


			for( int s = sampleframes ; s > 0 ; s-- )
			{

				CalculateIncrement( partial, pitch, mod );

				partial.IncrementPointer( oscillator.gate_state );

				float left, right;

				if( partial.IsStereo() )
				{
					assert( partial.s_ptr_l == partial.s_ptr_r + partial.s_ptr_l_offset );

					right = Interpolate( partial.s_ptr_r, partial.s_ptr_fine );
					left = Interpolate( partial.s_ptr_l, partial.s_ptr_fine );
				}
				else
				{
					if( PanningSupport == Panning )
					{
						float output = Interpolate( partial.s_ptr_r, partial.s_ptr_fine );
						left = output * partial.pan_left_level;
						right = output * partial.pan_right_level;
					}
					else
					{
						left = right = Interpolate( partial.s_ptr_r, partial.s_ptr_fine );

					}
				}

				if( addToOutput )
				{
					*output_l += left;
					*output_r += right;
				}
				else
				{
					*output_l = left;
					*output_r = right;
				}

				++output_l;
				++output_r;

			}

			if( partial.IsDone() )
			{
				it = oscillator.partials.erase( it );

				if( oscillator.partials.empty() )
				{
					//oscillator.NoteDone( bufferOffset + sampleframes - 1 ); // end note, origianal Jeff's code.
				    //oscillator.NoteDone( bufferOffset-1 ); // old kx's change
				    oscillator.NoteDone(-1 );
				}
			}
			else
			{
				++it;
			}

			// subsequent partials get summed.
			addToOutput = true;
		}
	};
};

#endif

